# -*- coding: utf-8 -*-
import logging
import math
import random
import time
from collections.abc import Iterator, Sized
from enum import Enum
from urllib.parse import parse_qs, urlsplit

import apistar
import requests

logger = logging.getLogger(__name__)


class PaginationMode(Enum):
    """Pagination modes with their different indexes"""

    PageNumber = "page"
    Cursor = "cursor"


class ResponsePaginator(Sized, Iterator):
    """
    A lazy generator to handle paginated Arkindex API endpoints.
    Does not perform any requests to the API until it is required.
    """

    def __init__(self, client, operation_id, *request_args, **request_kwargs):
        r"""
        :param client apistar.Client: An API client to use to perform requests for each page.
        :param \*request_args: Arguments to send to :meth:`apistar.Client.request`.
        :param \**request_kwargs: Keyword arguments to send to :meth:`apistar.Client.request`.
        """
        self.client = client
        """The APIStar client used to perform requests on each page."""

        self.data = {}
        """Stored data from the last performed request."""

        self.results = []
        """Stored results from the last performed request."""

        self.operation_id = operation_id
        """Client operation"""

        self.request_args = request_args
        """Arguments to send to :meth:`apistar.Client.request` with each request."""

        self.request_kwargs = request_kwargs
        """Keyword arguments to send to :meth:`apistar.Client.request` with each request."""

        self.mode = None
        """`page` for PageNumberPagination endpoints or `cursor` for CursorPagination endpoints."""

        self.count = None
        """Total results count."""

        self.pages_count = None
        """Total pages count."""

        self.pages_loaded = 0
        """Number of pages already loaded."""

        self.retries = request_kwargs.pop("retries", 5)
        assert (
            isinstance(self.retries, int) and self.retries > 0
        ), "retries must be a positive integer"
        """Max number of retries per API request"""

        # First page key is an empty string as we do not know yet the pagination type (e.g. page, cursor)
        self.initial_page = ""
        # Store retrieved pages remaining retries
        self.pages = {self.initial_page: self.retries}

        # Store missing page indexes
        self.missing = set()
        self.allow_missing_data = request_kwargs.pop("allow_missing_data", False)
        assert isinstance(
            self.allow_missing_data, bool
        ), "allow_missing_data must be a boolean"

    def _fetch_page(self):
        """
        Retrieve the next page and store its results

        Returns None in case of a failure
        Returns True in case the page returned results
        Returns False in case the page returned an empty result
        Raises a StopIteration in case there are no pages left to iterate on
        """
        # Filter out pages with no retries
        # Transform as a list of tuples for simpler output
        remaining = sorted([(m, v) for m, v in self.pages.items() if v > 0])

        # No remaining pages, end of iteration
        if not remaining:
            raise StopIteration

        # Get next page to load
        index, retry = remaining[0]

        if self.mode:
            self.request_kwargs[self.mode.value] = index

        try:
            extra_kwargs = {}
            if not self.pages_loaded:
                logger.info(
                    f"Loading first page on try {self.retries - retry + 1}/{self.retries}"
                )
                operation_fields = [
                    f.name
                    for f in self.client.lookup_operation(self.operation_id).fields
                ]
                # Ask to count results if the operation handle it as we do not know the pagination mode yet
                if "with_count" in operation_fields:
                    extra_kwargs["with_count"] = "true"
            else:
                logger.info(
                    f"Loading {self.mode.value} {index} on try {self.retries - retry + 1}/{self.retries}"
                    f" - remains {self.pages_count - self.pages_loaded} pages to load."
                )

            # Fetch the next page
            self.data = self.client.request(
                self.operation_id,
                *self.request_args,
                **self.request_kwargs,
                **extra_kwargs,
            )
            self.results = self.data.get("results", [])

            if not self.mode and self.data:
                # Autodetect if this endpoint uses page or cursor pagination
                if self.data.get("number"):
                    self.mode = PaginationMode.PageNumber
                else:
                    self.mode = PaginationMode.Cursor

            if self.count is None and "count" in self.data:
                # Retrieve information on first page with results count
                self.count = self.data["count"]
                if self.count == 0:
                    # Pagination has retrieved 0 results
                    self.pages = {}
                    return False
                self.pages_count = math.ceil(self.count / len(self.results))
                logger.info(
                    f"Pagination will load a total of {self.pages_count} pages."
                )
                if self.mode == PaginationMode.PageNumber:
                    # Initialize all pages once
                    self.pages = {
                        i: self.retries for i in range(2, self.pages_count + 1)
                    }
            elif self.mode == PaginationMode.PageNumber:
                # Mark page as loaded on other pages
                del self.pages[index]

            if self.mode == PaginationMode.Cursor:
                # Parse next URL to retrieve the cursor of the next page
                query = urlsplit(self.data["next"]).query
                cursor_query = parse_qs(query).get("cursor")
                if not cursor_query:
                    self.pages = {}
                else:
                    self.pages = {cursor_query[0]: self.retries}

            # Stop happy path here, we don't need to process errors
            self.pages_loaded += 1
            return True

        except apistar.exceptions.ErrorResponse as e:
            logger.warning(f"API Error {e.status_code} on pagination: {e.content}")

            # Decrement pages counter
            self.pages[index] -= 1

            # Sleep a bit (less than a second)
            time.sleep(random.random())

        except requests.exceptions.ConnectionError as e:
            logger.error(f"Server connection error, will retry in a few seconds: {e}")

            # Decrement pages counter
            self.pages[index] -= 1

            # Sleep a few seconds
            time.sleep(random.randint(1, 10))

        # Detect and store references to missing pages
        # when a page has no retries left
        if self.pages[index] <= 0:
            error_text = "No more retries left"
            if self.mode:
                error_text += f" for {self.mode.value} {index}"

            logger.warning(error_text)
            if self.allow_missing_data:
                self.missing.add(index)
            else:
                raise Exception("Stopping pagination as data will be incomplete")

        # No data could be fetch, return None
        return None

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.results) < 1:
            if self.data and self.data.get("next") is None:
                raise StopIteration

            # Continuously try to fetch a page until there are some retries left
            # This will still yield as soon as some data is fetched
            while self._fetch_page() is None:
                pass

        # Even after fetching a new page, if the new page is empty, just fail
        if len(self.results) < 1:
            raise StopIteration

        return self.results.pop(0)

    def __len__(self):
        # Handle calls to len when no requests have been made yet
        if not self.pages_loaded and self.count is None:
            # Continuously try to fetch a page until there are no retries left
            while self._fetch_page() is None:
                pass
        # Count may be null in case of a StopIteration
        if self.count is None:
            raise Exception("An error occurred fetching items total count")
        return self.count

    def __repr__(self):
        return "<{} via {!r}: {!r} {!r}>".format(
            self.__class__.__name__,
            self.client,
            (self.operation_id, self.request_args),
            self.request_kwargs,
        )



# -*- coding: utf-8 -*-
"""
Arkindex API Client
"""
import logging
import os
import warnings
from time import sleep
from urllib.parse import urljoin, urlsplit, urlunsplit

import apistar
import requests
import yaml

#from auth import TokenSessionAuthentication
from neuma_client.exceptions import NotFoundException
#from pagination import ResponsePaginator
#from transports import ArkindexHTTPTransport

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEFAULT_BASE_URL = "https://neuma.huma-num.fr/"

# Endpoint accessed by the client on instantiation to retrieve the OpenAPI schema
SCHEMA_ENDPOINT = "/api/v1/openapi/?format=json"

# Path to Neuma services
SERVICES_PATH = "rest"

logger = logging.getLogger(__name__)

# Create file handler
"""
f_handler = logging.FileHandler(__name__ + '.log')
f_handler.setLevel(logging.WARNING)
# Create formatters and add it to handlers
f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
f_handler.setFormatter(f_format)
# Add handlers to the logger
logger.addHandler(f_handler)
"""
# For the console
c_handler = logging.StreamHandler()
c_handler.setLevel(logging.WARNING)
c_format = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
c_handler.setFormatter(c_format)
logger.addHandler(c_handler)


def options_from_env():
	"""
	Get API client keyword arguments from environment variables.
	"""
	options = {}
	# ARKINDEX_TASK_TOKEN takes priority over ARKINDEX_API_TOKEN
	if "ARKINDEX_TASK_TOKEN" in os.environ:
		options["auth_scheme"] = "Ponos"
		options["token"] = os.environ.get("ARKINDEX_TASK_TOKEN")
	elif "ARKINDEX_API_TOKEN" in os.environ:
		options["auth_scheme"] = "Token"
		options["token"] = os.environ.get("ARKINDEX_API_TOKEN")

	# Allow overriding the default auth schemes
	if "ARKINDEX_API_AUTH_SCHEME" in os.environ:
		options["auth_scheme"] = os.environ.get("ARKINDEX_API_AUTH_SCHEME")

	if "ARKINDEX_API_URL" in os.environ:
		options["base_url"] = os.environ.get("ARKINDEX_API_URL")

	if "ARKINDEX_API_SCHEMA_URL" in os.environ:
		options["schema_url"] = os.environ.get("ARKINDEX_API_SCHEMA_URL")

	if "ARKINDEX_API_CSRF_COOKIE" in os.environ:
		options["csrf_cookie"] = os.environ.get("ARKINDEX_API_CSRF_COOKIE")

	return options


def _find_operation(schema, operation_id):
	for path_object in schema["paths"].values():
		for operation in path_object.values():
			if operation["operationId"] == operation_id:
				return operation
	raise KeyError("Operation '{}' not found".format(operation_id))


def _find_param(operation, param_name):
	for parameter in operation.get("parameters", []):
		if parameter["name"] == param_name:
			return parameter
	raise KeyError("Parameter '{}' not found".format(param_name))


class NeumaClient(apistar.Client):
	"""
	A Neuma API client.
	"""

	def __init__(
		self,
		token=None,
		auth_scheme="Token",
		base_url=DEFAULT_BASE_URL,
		schema_url=None,
		csrf_cookie=None,
		sleep=0,
		**kwargs,
	):
		r"""
		:param token: An API token to use. If omitted, access is restricted to public endpoints.
		:type token: str or None
		:param str base_url: A custom base URL for the client. If omitted, defaults to the Neuma main server.
		:param schema_url: URL or local path to an OpenAPI schema to use instead of the Arkindex instance's own schema.
		:type schema_url: str or None
		:param csrf_cookie: Use a custom CSRF cookie name. By default, the client will try to use any cookie
		   defined in ``x-csrf-cookie`` on the Server Object of the OpenAPI specification, and fall back to
		   ``arkindex.csrf``.
		:type csrf_cookie: str or None
		:param float sleep: Number of seconds to wait before sending each API request,
		   as a simple means of throttling.
		:param \**kwargs: Keyword arguments to send to ``apistar.Client``.
		"""
		if not schema_url:
			schema_url = urljoin(base_url, SCHEMA_ENDPOINT)

		try:
			split = urlsplit(schema_url)
			if split.scheme == "file" or not (split.scheme or split.netloc):
				# This is a local path
				with open(schema_url) as f:
					schema = yaml.safe_load(f)
			else:
				resp = requests.get(schema_url)
				resp.raise_for_status()
				schema = yaml.safe_load(resp.content)
		except Exception as e:
			raise SchemaError(
				f"Could not retrieve a proper OpenAPI schema from {schema_url}"
			) from e

		super().__init__(schema, **kwargs)

		# APIStar will treat a schema as valid even when there are no endpoints, making the client completely useless.
		if not len(self.document.walk_links()):
			raise SchemaError(
				f"The OpenAPI schema from {base_url} has no defined endpoints"
			)

		# Post-processing of the parsed schema
		for link_info in self.document.walk_links():
			# Look for deprecated links
			# https://github.com/encode/apistar/issues/664
			operation = _find_operation(schema, link_info.link.name)
			link_info.link.deprecated = operation.get("deprecated", False)
			for item in link_info.link.get_query_fields():
				parameter = _find_param(operation, item.name)
				item.deprecated = parameter.get("deprecated", False)

			# Detect paginated links
			if "x-paginated" in operation:
				link_info.link._paginated = operation["x-paginated"]

			# Remove domains from each endpoint; allows APIStar to properly handle our base URL
			# https://github.com/encode/apistar/issues/657
			original_url = urlsplit(link_info.link.url)
			# Removes the scheme and netloc
			new_url = ("", "", *original_url[2:])
			link_info.link.url = urlunsplit(new_url)

		# Try to autodetect the CSRF cookie:
		# - Try to find a matching server for this base URL and look for the x-csrf-cookie extension
		# - Fallback to arkindex.csrf
		if not csrf_cookie:
			split_base_url = urlsplit(base_url or self.document.url)
			csrf_cookies = {
				urlsplit(server.get("url", "")).netloc: server.get("x-csrf-cookie")
				for server in schema.get("servers", [])
			}
			csrf_cookie = csrf_cookies.get(split_base_url.netloc) or "arkindex.csrf"

		self.configure(
			token=token,
			auth_scheme=auth_scheme,
			base_url = base_url,
			services_url=  base_url, #urljoin(base_url, SERVICES_PATH),
			csrf_cookie=csrf_cookie,
			sleep=sleep,
		)

	def __repr__(self):
		return "<{} on {!r}>".format(
			self.__class__.__name__,
			self.document.url if hasattr(self, "document") else "",
		)

	def configure(
		self,
		token=None,
		auth_scheme="Token",
		base_url=None,
		services_url=None,
		csrf_cookie=None,
		sleep=None,
	):
		"""
		Reconfigure the API client.

		:param token: An API token to use. If omitted, access is restricted to public endpoints.
		:type token: str or None
		:param auth_scheme:
		   An authentication scheme to use. This is added in the HTTP header before the token:
		   ``Authorization: [scheme] [token]``.
		   This should use ``Token`` to authenticate as a regular user and ``Ponos`` to authenticate as a Ponos task.
		   If omitted, this defaults to ``Token``.
		:type auth_scheme: str or None
		:param base_url: The Neuma URL
		:type base_url: str or None
		:param services_url: A custom base URL for the client. If omitted, defaults to the Neuma main server.
		:type base_url: str or None
		:param csrf_cookie: Use a custom CSRF cookie name. Falls back to ``arkindex.csrf``.
		:type csrf_cookie: str or None
		:param float sleep: Number of seconds to wait before sending each API request,
		   as a simple means of throttling.
		"""
		
		'''
		if not csrf_cookie:
			csrf_cookie = "neuma.csrf"
			self.transport.session.auth = TokenSessionAuthentication(
			token,
			csrf_cookie_name=csrf_cookie,
			scheme=auth_scheme,
		)
		''' 
		if not sleep or not isinstance(sleep, float) or sleep < 0:
			self.sleep_duration = 0
		self.sleep_duration = sleep

		if base_url:
			self.base_url = base_url

		if services_url:
			self.document.url = services_url

		# Add the Referer header to allow Django CSRF to function
		self.transport.headers.setdefault("Referer", self.document.url)

	def url(self):
		return self.document.url
	
	def paginate(self, operation_id, *args, **kwargs):
		"""
		Perform a usual request as done by APIStar, but handle paginated endpoints.

		:return: An iterator for a paginated endpoint.
		:rtype: Union[arkindex.pagination.ResponsePaginator, dict, list]
		"""
		link = self.lookup_operation(operation_id)
		# If there was no x-paginated, trust the caller and assume the endpoint is paginated
		if getattr(link, "_paginated", True):
			return ResponsePaginator(self, operation_id, *args, **kwargs)
		return self.request(operation_id, *args, **kwargs)

	def login(self, email, password):
		"""
		Login to Neuma using an email/password combination.
		This helper method automatically sets the client's authentication settings with the token.
		"""
		resp = self.request("Login", body={"email": email, "password": password})
		if "auth_token" in resp:
			self.transport.session.auth.scheme = "Token"
			self.transport.session.auth.token = resp["auth_token"]
		return resp

	def request(self, operation_id, *args, **kwargs):
		"""
		Perform an API request.
		:param args: Arguments passed to the APIStar client.
		:param kwargs: Keyword arguments passed to the APIStar client.
		"""
		link = self.lookup_operation(operation_id)
		if link.deprecated:
			warnings.warn(
				"Endpoint '{}' is deprecated.".format(operation_id),
				DeprecationWarning,
				stacklevel=2,
			)

		query_params = self.get_query_params(link, kwargs)
		fields = link.get_query_fields()
		for field in fields:
			if field.deprecated and field.name in query_params:
				warnings.warn(
					"Parameter '{}' is deprecated.".format(field.name),
					DeprecationWarning,
					stacklevel=2,
				)
		if self.sleep_duration:
			logger.debug(
				"Delaying request by {:f} seconds...".format(self.sleep_duration)
			)
			sleep(self.sleep_duration)
		return super().request(operation_id, *args, **kwargs)



# -*- coding: utf-8 -*-
from apistar.client.auth import SessionAuthentication, TokenAuthentication


class TokenSessionAuthentication(TokenAuthentication):
    """
    A token authentication that takes care of CSRF tokens.
    """

    def __init__(
        self,
        token,
        scheme="Token",
        csrf_cookie_name="arkindex.csrf",
        csrf_header_name="X-CSRFToken",
    ):
        """
        :param str token: The API token to use.
        :param str scheme: The HTTP authentication scheme to use for token authentication.
        :param str csrf_cookie_name: Name of the CSRF token cookie.
        :param str csrf_header_name: Name of the CSRF request header.
        """
        self.session_auth = SessionAuthentication(
            csrf_cookie_name=csrf_cookie_name,
            csrf_header_name=csrf_header_name,
        )
        super().__init__(token, scheme=scheme)

    def __call__(self, request):
        request = self.session_auth(request)
        if self.token is None:
            return request
        return super().__call__(request)



"""
	A set of classes that provide proxy to Neuma object 
	through REST calls
"""

from neuma_client.exceptions import NotFoundException

from neuma_client.client import NeumaClient, logger

class Collections:
	"""
	Singleton object, the first that must be instantiated. Contains the Rest client
	"""

	def __init__(self, client) :
		self.client = client
		try:
			client.request ("NeumaApi_v3")
			logger.info (f"Connection OK to {client.url()}")
		except Exception as ex :
			logger.error (f"Unable to connect to {client.url()}: {ex}")

	def get_corpus(self, corpus_ref):
		try:
			res = self.client.request ("Element", full_neuma_ref=corpus_ref)
			return Corpus (self.client, res)
		except:
			raise NotFoundException (f"Corpus not found : {corpus_ref}")
			
class Corpus:

	def __init__(self, client, corpus_dict):		
		self.client = client
		self.ref = corpus_dict["ref"]
		self.title = corpus_dict["title"]
		self.short_title = corpus_dict["short_title"]
		self.description = corpus_dict["description"]
		self.short_description = corpus_dict["short_description"]
		self.parent = corpus_dict["parent"]
		self.composer = corpus_dict["composer"]
		self.licence = corpus_dict["licence"]
		self.copyright= corpus_dict["copyright"]
		
		self.opus_list = []
		self.opus_in_cache = False
		
	def get_opus_list(self):
		"""
			Find the opus in the corpus
		"""
		if not (self.opus_in_cache):
			res = self.client.request ("OpusList", full_neuma_ref=self.ref)
			for op_dict in res['results']:
				self.opus_list.append (Opus(self.client, op_dict))
			self.opus_in_cache = True
		return self.opus_list
	
	def __str__(self):
		return f"Corpus. ref:{self.ref}"

class Opus:

	def __init__(self, client, opus_dict):	
		self.client = client
		self.ref = opus_dict["ref"]	
		self.local_ref = opus_dict["local_ref"]	
		self.title = opus_dict["title"]	

		self.sources_in_cache = False 
		self.sources = []

	def get_sources_list(self):
		"""
			Find the sources in the opus
		"""
		if not (self.sources_in_cache):
			res = self.client.request ("SourceList", full_neuma_ref=self.ref)
			for src_dict in res['results']:
				self.sources .append (Source(self.client, self, src_dict))
			self.sources_in_cache = True
		return self.sources 

	def get_source (self, source_ref):
		if not (self.sources_in_cache):
			self.get_sources_list()
		for src in self.sources:
			if src.ref == source_ref:
				return src
		return None
			
	def source_exists (self, source_ref):
		if not (self.sources_in_cache):
			self.get_sources_list()
			
		for src in self.sources:
			if src.ref == source_ref:
				return True
		return False

class Source:

	IIIF_REF = "iiif"
	
	def __init__(self, client, opus, source_dict):	
		self.client = client
		self.opus = opus
		self.ref = source_dict["ref"]	
		self.url = source_dict["url"]	
		self.description = source_dict["description"]	
		self.source_type = source_dict["source_type"]	
		self.file_path = source_dict["file_path"]	
		self.has_manifest = source_dict["has_manifest"]	
		
		
		self.manifest = None
		self.manifest_in_cache = False
		
		
	def file_url(self):
		if self.file_path is not None:
			return self.client.base_url + self.file_path
		else:
			return None
	
	def __str__(self):
		return f"Source -- {self.opus.ref}/{self.ref}"
	
	"""
	    Part related to the manifest  -- if it exists
	"""
	def get_manifest(self):
		if self.has_manifest == False:
			return None
		
		if  self.manifest_in_cache == False:
			self.manifest  = self.client.request ("SourceManifest", 
											full_neuma_ref=self.opus.ref,
											source_ref=self.ref)
			self.manifest_in_cache = True
		return Manifest(self.manifest)


class Manifest:
	# Description of an IIIF source
	
	def __init__(self, manifest_dict):	
		# The manifest dictionary, obtained from the REST call
		self.manifest_dict  = manifest_dict

	def id_source(self):
		return self.manifest_dict ["id"]
	def url_source(self):
		return self.manifest_dict ["url"]
	
	def nb_pages(self):
		return len (self.manifest_dict["pages"])
	
	def get_page(self, i_page):
		if i_page >= len(self.manifest_dict["pages"]):
			raise Exception ("Attempt to read beyond the number of pages")		
		return self.manifest_dict["pages"][i_page]		
	def page_url (self, i_page):
		page = self.get_page(i_page)
		return page["url"]
	def page_width (self, i_page):
		page = self.get_page(i_page)
		return page["width"]
	def page_height (self, i_page):
		page = self.get_page(i_page)
		return page["height"]
	
	def nb_systems (self, i_page):
		page = self.get_page(i_page)
		return len(page["systems"])

	def first_music_page(self):
		if "first_music_page" in self.manifest_dict.keys():
			return self.manifest_dict["first_music_page"]
		else:
			return 1
	
	def get_system (self, i_page, i_system):
		page = self.get_page(i_page)
		if i_system >= len(page["systems"]):
			raise Exception (f"Attempt to read beyond the number of systems in page {i_page}")
		return page["systems"][i_system]
		
	def nb_staves(self, i_page, i_system):
		system = self.get_system(i_page, i_system)
		return len(system["staves"])
		
	def system_region(self, i_page, i_system):
		system = self.get_system(i_page, i_system)
		return system["region"]

	def nb_measures(self, i_page, i_system):
		system = self.get_system(i_page, i_system)
		if "measures" in system.keys():
			return len(system["measures"])
		else:
			raise Exception ("No 'measures' field in the manifest. Probably an obsolete version")		

	def get_measure(self, i_page, i_system, i_measure):
		system = self.get_system(i_page, i_system)
		if i_measure >= len(system["measures"]):
			raise Exception (f"Attempt to read beyond the number of measures in system {i_system}")
		return system["measures"][i_measure]
			



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

setup(
    packages=find_packages(
		 include=['neuma_client*'], 
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
	include_package_data=True
)
